@@ -1,7 +1,7 @@
 /*
  * BSD LICENSE
  *
- * Copyright(c) 2014-2017 Intel Corporation. All rights reserved.
+ * Copyright(c) 2014-2018 Intel Corporation. All rights reserved.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -51,12 +51,18 @@
 #include "profiles.h"
 #include "monitor.h"
 #include "alloc.h"
+#include "cap.h"
 
 /**
- * Default CDP configuration option - don't enforce on or off
+ * Default L3 CDP configuration option - don't enforce on or off
  */
 static enum pqos_cdp_config selfn_l3cdp_config = PQOS_REQUIRE_CDP_ANY;
 
+/**
+ * Default L2 CDP configuration option - don't enforce on or off
+ */
+static enum pqos_cdp_config selfn_l2cdp_config = PQOS_REQUIRE_CDP_ANY;
+
 /**
  * Monitoring reset
  */
@@ -92,10 +98,20 @@ static int sel_reset_alloc = 0;
  */
 static int sel_show_allocation_config = 0;
 
+/**
+ * Enable displaying supported RDT capabilities
+ */
+static int sel_display = 0;
+
+/**
+ * Enable displaying supported RDT capabilities in verbose mode
+ */
+static int sel_display_verbose = 0;
+
 /**
  * Selected library interface
  */
-int sel_interface = PQOS_INTER_MSR;
+enum pqos_interface sel_interface = PQOS_INTER_MSR;
 
 /**
  * @brief Function to check if a value is already contained in a table
@@ -293,26 +309,55 @@ selfn_super_verbose_mode(const char *arg)
 static void selfn_reset_alloc(const char *arg)
 {
         if (arg != NULL && (strlen(arg) > 0)) {
+                unsigned i;
+                char *tok = NULL;
+                char *saveptr = NULL;
+                char *s = NULL;
+
+                selfn_strdup(&s, arg);
+
                 const struct {
                         const char *name;
                         enum pqos_cdp_config cdp;
-                } patterns[] = {
+                } patternsl3[] = {
                         {"l3cdp-on",  PQOS_REQUIRE_CDP_ON},
                         {"l3cdp-off", PQOS_REQUIRE_CDP_OFF},
                         {"l3cdp-any", PQOS_REQUIRE_CDP_ANY},
+                }, patternsl2[] = {
+                        {"l2cdp-on",  PQOS_REQUIRE_CDP_ON},
+                        {"l2cdp-off", PQOS_REQUIRE_CDP_OFF},
+                        {"l2cdp-any", PQOS_REQUIRE_CDP_ANY},
                 };
-                unsigned i;
 
-                for (i = 0; i < DIM(patterns); i++)
-                        if (strcasecmp(arg, patterns[i].name) == 0)
-                                break;
+                tok = s;
+                while ((tok = strtok_r(tok, ",", &saveptr)) != NULL) {
+                        unsigned valid = 0;
 
-                if (i >= DIM(patterns)) {
-                        printf("Unrecognized '%s' allocation "
-                               "reset option!\n", arg);
-                        exit(EXIT_FAILURE);
+                        for (i = 0; i < DIM(patternsl3); i++)
+                                if (strcasecmp(tok, patternsl3[i].name) == 0) {
+                                        selfn_l3cdp_config = patternsl3[i].cdp;
+                                        valid = 1;
+                                        break;
+                                }
+
+                        for (i = 0; i < DIM(patternsl2); i++)
+                                if (strcasecmp(tok, patternsl2[i].name) == 0) {
+                                        selfn_l2cdp_config = patternsl2[i].cdp;
+                                        valid = 1;
+                                        break;
+                                }
+
+                        if (!valid) {
+                                printf("Unrecognized '%s' allocation "
+                                       "reset option!\n", tok);
+                                free(s);
+                                exit(EXIT_FAILURE);
+                        }
+
+                        tok = NULL;
                 }
-                selfn_l3cdp_config = patterns[i].cdp;
+
+                free(s);
         }
         sel_reset_alloc = 1;
 }
@@ -328,6 +373,28 @@ static void selfn_show_allocation(const char *arg)
         sel_show_allocation_config = 1;
 }
 
+/**
+ * @brief Selects displaying supported capabilities
+ *
+ * @param arg not used
+ */
+static void selfn_display(const char *arg)
+{
+        UNUSED_ARG(arg);
+        sel_display = 1;
+}
+
+/**
+ * @brief Selects displaying supported capabilities in verbose mode
+ *
+ * @param arg not used
+ */
+static void selfn_display_verbose(const char *arg)
+{
+        UNUSED_ARG(arg);
+        sel_display_verbose = 1;
+}
+
 /**
  * @brief Selects allocation profile from internal DB
  *
@@ -367,6 +434,8 @@ parse_config_file(const char *fname)
                 void (*fn)(const char *);
         } optab[] = {
                 {"show-alloc:",         selfn_show_allocation },   /**< -s */
+                {"display:",            selfn_display },           /**< -d */
+                {"display-verbose:",    selfn_display_verbose },   /**< -D */
                 {"log-file:",           selfn_log_file },          /**< -l */
                 {"verbose-mode:",       selfn_verbose_mode },      /**< -v */
                 {"super-verbose-mode:", selfn_super_verbose_mode },/**< -V */
@@ -450,8 +519,9 @@ static const char help_printf_short[] =
         "Usage: %s [-h] [--help] [-v] [--verbose] [-V] [--super-verbose]\n"
         "          [-l FILE] [--log-file=FILE] [-I] [--iface-os]\n"
         "       %s [-s] [--show]\n"
-        "       %s [-m EVTCORES] [--mon-core=EVTCORES] | [-p EVTPIDS] "
-        "[--mon-pid=EVTPIDS]\n"
+        "       %s [-d] [--display] [-D] [--display-verbose]\n"
+        "       %s [-m EVTCORES] [--mon-core=EVTCORES] | [-p [EVTPIDS]] "
+        "[--mon-pid[=EVTPIDS]]\n"
         "          [-t SECONDS] [--mon-time=SECONDS]\n"
         "          [-i N] [--mon-interval=N]\n"
         "          [-T] [--mon-top]\n"
@@ -471,6 +541,8 @@ static const char help_printf_long[] =
         "  -v, --verbose               verbose mode\n"
         "  -V, --super-verbose         super-verbose mode\n"
         "  -s, --show                  show current PQoS configuration\n"
+        "  -d, --display               display supported capabilities\n"
+        "  -D, --display-verbose       display supported capabilities in verbose mode\n"
         "  -f FILE, --config-file=FILE load commands from selected file\n"
         "  -l FILE, --log-file=FILE    log messages into selected file\n"
         "  -e CLASSDEF, --alloc-class=CLASSDEF\n"
@@ -480,6 +552,7 @@ static const char help_printf_long[] =
         "          Examples: 'llc:0=0xffff;llc:1=0x00ff;llc@0-1:2=0xff00',\n"
 	"                    'llc:0d=0xfff;llc:0c=0xfff00',\n"
         "                    'l2:2=0x3f;l2@2:1=0xf',\n"
+        "                    'l2:2d=0xf;l2:2c=0xc,\n"
         "                    'mba:1=30;mba@1:3=80'.\n"
         "  -a CLASS2ID, --alloc-assoc=CLASS2ID\n"
         "          associate cores/tasks with an allocation class.\n"
@@ -487,21 +560,28 @@ static const char help_printf_long[] =
         "          Example 'llc:0=0,2,4,6-10;llc:1=1'.\n"
         "          Example 'core:0=0,2,4,6-10;core:1=1'.\n"
         "          Example 'pid:0=3543,7643,4556;pid:1=7644'.\n"
-        "  -R [CONFIG], --alloc-reset[=CONFIG]\n"
+        "  -R [CONFIG[,CONFIG]], --alloc-reset[=CONFIG[,CONFIG]]\n"
         "          reset allocation configuration (L2/L3 CAT & MBA)\n"
-        "          CONFIG can be: l3cdp-on, l3cdp-off or l3cdp-any (default).\n"
+        "          CONFIG can be: l3cdp-on, l3cdp-off, l3cdp-any,\n"
+        "                         l2cdp-on, l2cdp-off, l2cdp-any\n"
+        "          (default l3cdp-any,l2cdp-any).\n"
         "  -m EVTCORES, --mon-core=EVTCORES\n"
         "          select cores and events for monitoring.\n"
         "          EVTCORES format is 'EVENT:CORE_LIST'.\n"
         "          Example: \"all:0,2,4-10;llc:1,3;mbr:11-12\".\n"
         "          Cores can be grouped by enclosing them in square brackets,\n"
         "          example: \"llc:[0-3];all:[4,5,6];mbr:[0-3],7,8\".\n"
-        "  -p EVTPIDS, --mon-pid=EVTPIDS\n"
-        "          select process ids and events to monitor.\n"
+        "  -p [EVTPIDS], --mon-pid[=EVTPIDS]\n"
+        "          select top 10 most active (CPU utilizing) process ids to monitor\n"
+        "          or select process ids and events to monitor.\n"
         "          EVTPIDS format is 'EVENT:PID_LIST'.\n"
-        "          Example 'llc:22,25673' or 'all:892,4588-4592'.\n"
-        "          Note: processes and cores cannot be monitored together.\n"
-        "                Requires Linux and kernel versions 4.1 and newer.\n"
+        "          Examples: 'llc:22,25673' or 'all:892,4588-4592'\n"
+        "          Process' IDs can be grouped by enclosing them in square brackets,\n"
+        "          Examples: 'llc:[22,25673]' or 'all:892,[4588-4592]'\n"
+        "          Note:\n"
+        "               Requires Linux and kernel versions 4.10 and newer.\n"
+        "               The -I option must be used for PID monitoring.\n"
+        "               Processes and cores cannot be monitored together.\n"
         "  -o FILE, --mon-file=FILE    output monitored data in a FILE\n"
         "  -u TYPE, --mon-file-type=TYPE\n"
         "          select output file format type for monitored data.\n"
@@ -531,33 +611,35 @@ static const char help_printf_long[] =
 static void print_help(const int is_long)
 {
         printf(help_printf_short,
-               m_cmd_name, m_cmd_name, m_cmd_name, m_cmd_name,
+               m_cmd_name, m_cmd_name, m_cmd_name, m_cmd_name, m_cmd_name,
                m_cmd_name, m_cmd_name, m_cmd_name);
         if (is_long)
-                printf(help_printf_long);
+                printf("%s", help_printf_long);
 }
 
 static struct option long_cmd_opts[] = {
-        {"help",          no_argument,       0, 'h'},
-        {"log-file",      required_argument, 0, 'l'},
-        {"config-file",   required_argument, 0, 'f'},
-        {"show",          no_argument,       0, 's'},
-        {"profile-list",  no_argument,       0, 'H'},
-        {"profile-set",   required_argument, 0, 'c'},
-        {"mon-interval",  required_argument, 0, 'i'},
-        {"mon-pid",       required_argument, 0, 'p'},
-        {"mon-core",      required_argument, 0, 'm'},
-        {"mon-time",      required_argument, 0, 't'},
-        {"mon-top",       no_argument,       0, 'T'},
-        {"mon-file",      required_argument, 0, 'o'},
-        {"mon-file-type", required_argument, 0, 'u'},
-        {"mon-reset",     no_argument,       0, 'r'},
-        {"alloc-class",   required_argument, 0, 'e'},
-        {"alloc-reset",   required_argument, 0, 'R'},
-        {"alloc-assoc",   required_argument, 0, 'a'},
-        {"verbose",       no_argument,       0, 'v'},
-        {"super-verbose", no_argument,       0, 'V'},
-        {"iface-os",      no_argument,       0, 'I'},
+        {"help",            no_argument,       0, 'h'},
+        {"log-file",        required_argument, 0, 'l'},
+        {"config-file",     required_argument, 0, 'f'},
+        {"show",            no_argument,       0, 's'},
+        {"display",         no_argument,       0, 'd'},
+        {"display-verbose", no_argument,       0, 'D'},
+        {"profile-list",    no_argument,       0, 'H'},
+        {"profile-set",     required_argument, 0, 'c'},
+        {"mon-interval",    required_argument, 0, 'i'},
+        {"mon-pid",         required_argument, 0, 'p'},
+        {"mon-core",        required_argument, 0, 'm'},
+        {"mon-time",        required_argument, 0, 't'},
+        {"mon-top",         no_argument,       0, 'T'},
+        {"mon-file",        required_argument, 0, 'o'},
+        {"mon-file-type",   required_argument, 0, 'u'},
+        {"mon-reset",       no_argument,       0, 'r'},
+        {"alloc-class",     required_argument, 0, 'e'},
+        {"alloc-reset",     required_argument, 0, 'R'},
+        {"alloc-assoc",     required_argument, 0, 'a'},
+        {"verbose",         no_argument,       0, 'v'},
+        {"super-verbose",   no_argument,       0, 'V'},
+        {"iface-os",        no_argument,       0, 'I'},
         {0, 0, 0, 0} /* end */
 };
 
@@ -578,7 +660,7 @@ int main(int argc, char **argv)
         memset(&cfg, 0, sizeof(cfg));
 
         while ((cmd = getopt_long(argc, argv,
-                                  ":Hhf:i:m:Tt:l:o:u:e:c:a:p:srvVIR:",
+                                  ":Hhf:i:m:Tt:l:o:u:e:c:a:p:sdDrvVIR:",
                                   long_cmd_opts, &opt_index)) != -1) {
                 switch (cmd) {
                 case 'h':
@@ -600,7 +682,20 @@ int main(int argc, char **argv)
                         selfn_monitor_interval(optarg);
                         break;
                 case 'p':
-		        selfn_monitor_pids(optarg);
+                        if (optarg != NULL && *optarg == '-') {
+                                /**
+                                 * Next switch option wrongly assumed to be
+                                 * argument to '-p'.
+                                 * In order to fix it, we are handling this as
+                                 * '-p' without parameters (as it should be)
+                                 * to start top-pids monitoring mode.
+                                 * Have to rewind \a optind as well.
+                                 */
+                                selfn_monitor_top_pids();
+                                optind--;
+                                break;
+                        }
+                        selfn_monitor_pids(optarg);
                         pid_flag = 1;
                         break;
                 case 'm':
@@ -628,32 +723,42 @@ int main(int argc, char **argv)
                         sel_mon_reset = 1;
                         break;
                 case 'R':
-                        if (optarg != NULL && *optarg == '-') {
-                                /**
-                                 * Next switch option wrongly assumed to be
-                                 * argument to '-R'.
-                                 * Pass NULL as argument to '-R' function and
-                                 * rewind \a optind.
-                                 */
+                        if (optarg != NULL) {
+                                if (*optarg == '-') {
+                                        /**
+                                        * Next switch option wrongly assumed
+                                        * to be argument to '-R'.
+                                        * Pass NULL as argument to '-R' function
+                                        * and rewind \a optind.
+                                        */
+                                        selfn_reset_alloc(NULL);
+                                        optind--;
+                                } else
+                                        selfn_reset_alloc(optarg);
+                        } else
                                 selfn_reset_alloc(NULL);
-                                optind--;
-                        } else {
-                                selfn_reset_alloc(optarg);
-                        }
                         break;
                 case ':':
                         /**
                          * This is handler for missing mandatory argument
                          * (enabled by leading ':' in getopt() argument).
-                         * -R is only allowed switch for optional argument.
+                         * -R and -p are only allowed switch for optional args.
                          * Other switches need to report error.
                          */
-                        if (optopt != 'R') {
+                        if (optopt == 'R') {
+                                selfn_reset_alloc(NULL);
+                        } else if (optopt == 'p') {
+                                /**
+                                 * Top pids mode - in case of '-I -p' top N
+                                 * pids (by CPU usage) will be displayed and
+                                 * monitored for cache/mbm/misses
+                                 */
+                                selfn_monitor_top_pids();
+                                pid_flag = 1;
+                        } else {
                                 printf("Option -%c is missing required "
                                        "argument\n", optopt);
                                 return EXIT_FAILURE;
-                        } else {
-                                selfn_reset_alloc(NULL);
                         }
                         break;
                 case 'a':
@@ -666,6 +771,12 @@ int main(int argc, char **argv)
                 case 's':
                         selfn_show_allocation(NULL);
                         break;
+                case 'd':
+                        selfn_display(NULL);
+                        break;
+                case 'D':
+                        selfn_display_verbose(NULL);
+                        break;
                 case 'v':
                         selfn_verbose_mode(NULL);
                         break;
@@ -760,7 +871,14 @@ int main(int argc, char **argv)
         }
 
         if (sel_mon_reset && cap_mon != NULL) {
-                if (pqos_mon_reset() != PQOS_RETVAL_OK) {
+                ret = pqos_mon_reset();
+                if (sel_interface != PQOS_INTER_MSR &&
+                    ret == PQOS_RETVAL_RESOURCE) {
+                        exit_val = EXIT_FAILURE;
+                        printf("Monitoring cannot be reset on systems "
+                               "without resctrl monitoring capability. "
+                               "Required kernel version 4.14 or newer.\n");
+                } else if (ret != PQOS_RETVAL_OK) {
                         exit_val = EXIT_FAILURE;
                         printf("CMT/MBM reset failed!\n");
                 } else {
@@ -772,7 +890,9 @@ int main(int argc, char **argv)
                 /**
                  * Reset allocation configuration to after-reset state and exit
                  */
-                if (pqos_alloc_reset(selfn_l3cdp_config) != PQOS_RETVAL_OK) {
+                ret = pqos_alloc_reset(selfn_l3cdp_config,
+                                       selfn_l2cdp_config);
+                if (ret != PQOS_RETVAL_OK) {
                         exit_val = EXIT_FAILURE;
                         printf("Allocation reset failed!\n");
                 } else
@@ -789,6 +909,14 @@ int main(int argc, char **argv)
                 goto allocation_exit;
         }
 
+        if (sel_display || sel_display_verbose) {
+                /**
+                 * Display info about supported capabilities
+                 */
+                cap_print_features(p_cap, p_cpu, sel_display_verbose);
+                goto allocation_exit;
+        }
+
         if (sel_allocation_profile != NULL) {
                 if (profile_l3ca_apply(sel_allocation_profile, cap_l3ca) != 0) {
                         exit_val = EXIT_FAILURE;
